#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/time.h>
#include <math.h>
#include <pthread.h>

#include <CL/cl.h>


#define MAX_PLATFORMS 16
#define MAX_DEVICES 4


int size, slice, tile;
char **kernel_source_code;
struct{
	cl_platform_id platform;
	cl_device_id device;
	size_t resolution;
	size_t total_slices;
	size_t total_enq_write_bytes;
	size_t total_enq_read_bytes;
	cl_ulong total_kernel_work;
} dev_info[MAX_PLATFORMS*MAX_DEVICES];
float *A, *B, *C;
pthread_t dev_thread[MAX_PLATFORMS*MAX_DEVICES];
pthread_mutex_t get_work_mutex=PTHREAD_MUTEX_INITIALIZER;
pthread_barrier_t _ccl_barrier;


char ** read_program_sources(int nfiles, ...){
	FILE *file;
	long size;
	va_list ap;
	int n;
	char **strings;

	strings = malloc(sizeof(char*) * nfiles);
	va_start(ap, nfiles);
	for(n=0; n<nfiles; n++){
		file = fopen(va_arg(ap, char*), "r");
		fseek(file, 0, SEEK_END);
		size = ftell(file);
		rewind(file);
		(strings)[n] = (char *)malloc(sizeof(char) * size + 1);
		fread((strings)[n], 1, sizeof(char) * size, file);
		(strings)[n][size] = '\0';
		fclose(file);
	}
	va_end(ap);

	return(strings);
}


void test_cl_success(char *where, cl_int error){
	char error_name[65][44]={"CL_SUCCESS", "CL_DEVICE_NOT_FOUND", "CL_DEVICE_NOT_AVAILABLE", "CL_COMPILER_NOT_AVAILABLE", "CL_MEM_OBJECT_ALLOCATION_FAILURE", "CL_OUT_OF_RESOURCES", "CL_OUT_OF_HOST_MEMORY", "CL_PROFILING_INFO_NOT_AVAILABLE", "CL_MEM_COPY_OVERLAP", "CL_IMAGE_FORMAT_MISMATCH", "CL_IMAGE_FORMAT_NOT_SUPPORTED", "CL_BUILD_PROGRAM_FAILURE", "CL_MAP_FAILURE", "CL_MISALIGNED_SUB_BUFFER_OFFSET", "CL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "CL_INVALID_VALUE", "CL_INVALID_DEVICE_TYPE", "CL_INVALID_PLATFORM", "CL_INVALID_DEVICE", "CL_INVALID_CONTEXT",  "CL_INVALID_QUEUE_PROPERTIES", "CL_INVALID_COMMAND_QUEUE", "CL_INVALID_HOST_PTR", "CL_INVALID_MEM_OBJECT", "CL_INVALID_IMAGE_FORMAT_DESCRIPTOR", "CL_INVALID_IMAGE_size", "CL_INVALID_SAMPLER", "CL_INVALID_BINARY", "CL_INVALID_BUILD_OPTIONS", "CL_INVALID_PROGRAM", "CL_INVALID_PROGRAM_EXECUTABLE", "CL_INVALID_KERNEL_NAME", "CL_INVALID_KERNEL_DEFINITION", "CL_INVALID_KERNEL", "CL_INVALID_ARG_INDEX", "CL_INVALID_ARG_VALUE", "CL_INVALID_ARG_size", "CL_INVALID_KERNEL_ARGS", "CL_INVALID_WORK_DIMENSION", "CL_INVALID_WORK_GROUP_size", "CL_INVALID_WORK_ITEM_size", "CL_INVALID_GLOBAL_OFFSET", "CL_INVALID_EVENT_WAIT_LIST", "CL_INVALID_EVENT", "CL_INVALID_OPERATION", "CL_INVALID_GL_OBJECT", "CL_INVALID_BUFFER_size", "CL_INVALID_MIP_LEVEL", "CL_INVALID_GLOBAL_WORK_size", "CL_INVALID_PROPERTY"};

	if(error != CL_SUCCESS){
		printf("%s: %s\n", where, error_name[-error]);
		exit(1);
	}
}


void perror_and_exit(char *str){
	perror(str);
	exit(1);
}


void get_work(int *i, int *j){
	static int ii=0, jj=0;
	
	pthread_mutex_lock(&get_work_mutex);

	*i = ii;
	*j = jj;
	if((ii != -1) && (jj != -1)){
		jj += slice;
		if(jj == size){
			ii += slice;
			jj = 0;
		}
		if(ii == size)
			ii = jj = -1;
	}

	pthread_mutex_unlock(&get_work_mutex);
}


void * dev_thread_routine(void *arg){
	cl_context_properties properties[3]={CL_CONTEXT_PLATFORM, 0, 0};
	cl_context ctx;
	cl_command_queue command_queue;
	cl_mem bufferA, bufferB, bufferC;
	cl_program program;
	cl_kernel kernel;
	cl_event event[3];
	cl_int result, errcode_ret;
	cl_ulong c_start, c_end;
	size_t global_work_size[2], local_work_size[2];
	int dev, i, j, last_i=-1, last_j=-1, i2, j2;
	float *C2;

	pthread_barrier_wait(&_ccl_barrier);
	
	C2 = (float *)malloc(sizeof(float)*slice*slice);

	dev = (long)((int *)arg);
	properties[1] = (cl_context_properties)dev_info[dev].platform;
	ctx = clCreateContext(properties, 1, &dev_info[dev].device, NULL, NULL, &errcode_ret);
	test_cl_success("clCreateContext", errcode_ret);

	command_queue = clCreateCommandQueue(ctx, dev_info[dev].device, CL_QUEUE_PROFILING_ENABLE
, &errcode_ret);
	test_cl_success("clCreateCommandQueue", errcode_ret);

	bufferA = clCreateBuffer(ctx, CL_MEM_READ_ONLY, sizeof(float)*size*slice, NULL, &errcode_ret);
	test_cl_success("clCreateBuffer", errcode_ret);
	bufferB = clCreateBuffer(ctx, CL_MEM_READ_ONLY, sizeof(float)*slice*size, NULL, &errcode_ret);
	test_cl_success("clCreateBuffer", errcode_ret);
	bufferC = clCreateBuffer(ctx, CL_MEM_WRITE_ONLY, sizeof(float)*slice*slice, NULL, &errcode_ret);
	test_cl_success("clCreateBuffer", errcode_ret);

	program = clCreateProgramWithSource(ctx, 1, (const char **)kernel_source_code, NULL, &errcode_ret);
	test_cl_success("clCreateProgramWithSource", errcode_ret);

	result = clBuildProgram(program, 1, &dev_info[dev].device, "", NULL, NULL);
	test_cl_success("clBuildProgram", result);

	kernel = clCreateKernel(program, "matrix_mult", &errcode_ret);
	test_cl_success("clCreateKernel", errcode_ret);

	result = clSetKernelArg(kernel, 0, sizeof(int), &size);
	test_cl_success("clSetKernelArg", result);
	result = clSetKernelArg(kernel, 1, sizeof(int), &slice);
	test_cl_success("clSetKernelArg", result);
	result = clSetKernelArg(kernel, 2, sizeof(cl_mem), &bufferA);
	test_cl_success("clSetKernelArg", result);
	result = clSetKernelArg(kernel, 3, sizeof(cl_mem), &bufferB);
	test_cl_success("clSetKernelArg", result);
	result = clSetKernelArg(kernel, 4, sizeof(cl_mem), &bufferC);
	test_cl_success("clSetKernelArg", result);

	if(tile == 0)
		global_work_size[0] = global_work_size[1] = local_work_size[0] = local_work_size[1] = 1;
	else{
		global_work_size[0] = global_work_size[1] = slice;
		local_work_size[0] = local_work_size[1] = tile;
	}
	
	get_work(&i, &j);
	while((i != -1) && (j != -1)){
		printf("[%04d-%04d][%04d-%04d]: dev=%02d, platform=%#04lx, device=%#04lx\n", i, i+slice-1, j, j+slice-1, dev, (long)dev_info[dev].platform, (long)dev_info[dev].device);
		dev_info[dev].total_slices++;

		if(i != last_i){
			result = clEnqueueWriteBuffer(command_queue, bufferA, CL_FALSE, 0, sizeof(float)*size*slice, A+size*i, 0, NULL, &event[0]);
			test_cl_success("clEnqueueWriteBuffer", result);
			dev_info[dev].total_enq_write_bytes += sizeof(float)*size*slice;
		}

		if(j != last_j){
			result = clEnqueueWriteBuffer(command_queue, bufferB, CL_FALSE, 0, sizeof(float)*slice*size, B+size*j, 0, NULL, &event[1]);
			test_cl_success("clEnqueueWriteBuffer", result);
			dev_info[dev].total_enq_write_bytes += sizeof(float)*size*size;
		}

		result = clEnqueueNDRangeKernel(command_queue, kernel, 2, NULL, global_work_size, local_work_size, 2, event, &event[2]);
		test_cl_success("clEnqueueNDRangeKernel", result);

		result = clEnqueueReadBuffer(command_queue, bufferC, CL_TRUE, 0, sizeof(float)*slice*slice, C2, 1, &event[2], NULL);
		test_cl_success("clEnqueueReadBuffer", result);
		dev_info[dev].total_enq_read_bytes += sizeof(float)*slice*slice;

		result = clGetEventProfilingInfo(event[2], CL_PROFILING_COMMAND_START, sizeof(cl_ulong), &c_start, NULL);
		test_cl_success("clGetEventProfilingInfo", result);
		result = clGetEventProfilingInfo(event[2], CL_PROFILING_COMMAND_END, sizeof(cl_ulong), &c_end, NULL);
		test_cl_success("clGetEventProfilingInfo", result);
		dev_info[dev].total_kernel_work += c_end - c_start;

		for(i2=0; i2<slice; i2++)
			for(j2=0; j2<slice; j2++)
				C[(i+i2)*size+j+j2] = C2[i2*slice+j2];
		
		last_i = i;
		last_j = j;
		get_work(&i, &j);
	}

	pthread_exit(NULL);
}


void list_devices(){
	cl_uint num_platforms, num_devices;
	cl_platform_id platform[MAX_PLATFORMS];
	cl_device_id device[MAX_DEVICES];
	char platform_name[256], hostname[128], device_name[256];
	size_t param_size;
	cl_int result;
	int p, d;
	
	result = clGetPlatformIDs(MAX_PLATFORMS, platform, &num_platforms);
	test_cl_success("clGetPlatformIDs", result);

	for(p=0; p<num_platforms; p++){
		result = clGetPlatformInfo(platform[p], CL_PLATFORM_NAME, sizeof(platform_name), platform_name, &param_size);
		result = clGetPlatformInfo(platform[p], 0x000, sizeof(hostname), hostname, &param_size);
		printf("[%d] %s (%s)\n", p, platform_name, hostname);

		result = clGetDeviceIDs(platform[p], CL_DEVICE_TYPE_ALL, MAX_DEVICES, device, &num_devices);
		test_cl_success("clGetDeviceIDs", result);
		
		for(d=0; d<num_devices; d++){
			result = clGetDeviceInfo(device[d], CL_DEVICE_NAME, sizeof(device_name), &device_name, &param_size);
			printf("\t[%d] %s\n", d, device_name);
		}
	}
}


int main (int argc, char *argv[]){
	cl_uint num_platforms, num_devices;
	cl_platform_id platform[MAX_PLATFORMS];
	cl_device_id device[MAX_DEVICES];
	cl_int result;
	int p, d, num_devs=0, i, j, num_threads=0;
	float v;
	long double acc_errors=0;
	struct timeval t0, t1;
	unsigned long long tu0, tu1;

	if(argc != 5){
		printf("Usage: %s size slice tile\n", argv[0]);
		list_devices();
		exit(1);
	}
	size = atoi(argv[1]);
	slice = atoi(argv[2]);
	tile = atoi(argv[3]);
		
	A = (float *)malloc(sizeof(float)*size*size);
	B = (float *)malloc(sizeof(float)*size*size);
	C = (float *)malloc(sizeof(float)*size*size);
	for(i=0; i<size; i++)
		for(j=0; j<size; j++){
			A[i*size+j] = i+1;//stored by rows
			B[j*size+i] = j+2;//stored by columns
		}

	gettimeofday(&t0, NULL);

	kernel_source_code = read_program_sources(1, "cl_matrix_kernel.c");

	result = clGetPlatformIDs(MAX_PLATFORMS, platform, &num_platforms);
	test_cl_success("clGetPlatformIDs", result);

	for(p=0; p<num_platforms; p++){
		result = clGetDeviceIDs(platform[p], CL_DEVICE_TYPE_ALL, MAX_DEVICES, device, &num_devices);
		test_cl_success("clGetDeviceIDs", result);

		for(d=0; d<num_devices; d++){
			dev_info[num_devs].platform = platform[p];
			dev_info[num_devs].device = device[d];
			result = clGetDeviceInfo(dev_info[num_devs].device, CL_DEVICE_PROFILING_TIMER_RESOLUTION, sizeof(size_t), &dev_info[num_devs].resolution, NULL);
			test_cl_success("clGetDeviceInfo", result);
			dev_info[num_devs].total_slices = dev_info[num_devs].total_enq_write_bytes = dev_info[num_devs].total_enq_read_bytes = dev_info[num_devs].total_kernel_work = 0;			
			if(argv[4][num_devs] == 'Y')
				num_threads++;
			num_devs++;
		}
	}
	
	if(pthread_barrier_init(&_ccl_barrier, NULL, num_threads) != 0)
		perror_and_exit("pthread_barrier_init");	
	for(i=0; i<num_devs; i++)
		if(argv[4][i] == 'Y')
			if(pthread_create(&dev_thread[i], NULL, dev_thread_routine, (int *)(long)i) != 0)
				perror_and_exit("pthread_create");
	for(i=0; i<num_devs; i++)
		if(argv[4][i] == 'Y')
			pthread_join(dev_thread[i], NULL);
		
	gettimeofday(&t1, NULL);
	
	for(i=0; i<size; i+=64){
		v = 0;
		for(j=0; j<size; j++)
			v += A[i*size+j] * B[i*size+j];
		if(C[i*size+i] != v){
			acc_errors += fabs(C[i*size+i] - v);
			//printf("C[%d][%d]: %f; %f\n", i, i, C[i*size+i], v);
		}
	}
	printf("acc_errors=%LG\n", acc_errors);

	tu0 = (unsigned long long)t0.tv_sec*1000000L+(unsigned long long)t0.tv_usec;
	tu1 = (unsigned long long)t1.tv_sec*1000000L+(unsigned long long)t1.tv_usec;
	printf("time=%.3lfs\n", (tu1 - tu0) / 1000000.0);

	for(i=0; i<num_devs; i++)
		if(argv[4][i] == 'Y')
			//printf("dev %02d: total_slices=%ld, total_enq_wite_bytes=%ldM, total_enq_read_bytes=%ldM, total_kernel_load=%.3lfs\n", i, dev_info[i].total_slices, dev_info[i].total_enq_write_bytes / 1024 / 1024, dev_info[i].total_enq_read_bytes / 1024 / 1024, (dev_info[i].total_kernel_work * dev_info[i].resolution) / 1000000000.0);
			printf("dev %02d: total_slices=%ld, total_enq_wite_bytes=%ldM, total_enq_read_bytes=%ldM, total_kernel_load=%.3lfs\n", i, dev_info[i].total_slices, dev_info[i].total_enq_write_bytes / 1024 / 1024, dev_info[i].total_enq_read_bytes / 1024 / 1024, (double)dev_info[i].total_kernel_work / (double)1000000000.0);
		
	exit(0);
}
