
/*
__kernel void matrix_mult(const int size, const int slice, __global float *A, __global float *B, __global float *C){
	int i, j, k;
	float v=0;
	
	i = get_global_id(0);
	j = get_global_id(1);
	for(k=0; k<size; k++)
		v += A[i*size+k] * B[j*size+k];

	C[i*slice+j] = v;
}
*/

/*
// OpenCL Kernel for BLOCK matrix multiply. C = A * B
__kernel void matrix_mult(int wA, int wB, __global float *A, __global float *B, __global float *C) {

	int tx = get_global_id (0) ; // 2D Thread ID x
	int ty = get_global_id (1) ; // 2D Thread ID y
	
	// Perform dot-product and accumulate into value
	float value = 0 ;
	for (int k = 0 ; k < wA; ++k ) {
		value += A[ ty * wA + k ] * B[ k * wB + tx ] ; // valido se B nao vier transposta
		//value += A[ ty * wA + k ] * B[ tx * wA + k ] ;   // valido se B ja estiver transposta
	}
	C[ ty * wA + tx ] = value ; // Write to device memory; valido se B nao vier transposta
	//C[ ty * wB + tx ] = value; // Write to device memory; valido se B ja estiver transposta
}
*/




 // - global_work_size = array 2D que define o número total de work-items (threads), de uma "grelha global de threads"
 //                      no nosso caso, global_work_size[0] = global_work_size[1] = GLOBAL_slice
 //                      (cada device irá produzir uma sub-matriz de C, de ordem GLOBAL_slice x GLOBAL_slice),
 //                      donde, em cada device executarão um total de GLOBAL_slice^2 threads
 // - work-item = thread da grelha global de threads, com índice único no array 2D global_work_size
 // - work-group = sub-array 2D da grelha global de threads, em que os threads têm capacidade de sincronizacao
 // - work-group-size = array 2D que define o número de threads de um work-group;
 //                     no nosso caso, local_work_size[0] = local_work_size[1] = GLOBAL_tile;
 // 			        global_work_size tem de ser multiplo do work-group-size;
 //                     (no nosso caso, o parametro slice tem de se multiplo do parametro tile);
 //                     tile nunca pode ser superior a 32=2^5 pois 2^5*2^5=2^10=1024=CL_DEVICE_MAX_WORK_GROUP_SIZE
 
 // GTX 460:
 // - CL_DEVICE_MAX_WORK_GROUP_SIZE:	1024
 // - CL_DEVICE_WARP_SIZE_NV:		32
 // -- por cada work-group pode executar um warp-group de 32 threads em simultaneo;
 //    (se um work-group tiver vários warp-groups, estes serã serializadas)
 // - CL_DEVICE_LOCAL_MEM_SIZE:		48 KByte
 // -- que isto dizer que a memória local total nuca pode ultrapassar 48K
 //    como K = 2^5 x 2^5 e um float são 4 bytes, então 


// OpenCL Kernel for BLOCK matrix multiply. C = A * B
#define BLOCK_SIZE 16

__kernel void matrix_mult(int wA, int wB, __global float *A, __global float *B, __global float *C) {
	
	int bx = get_group_id (0); // 2D Thread ID x
	int by = get_group_id (1); // 2D Thread ID y

	//const int BLOCK_SIZE = 16;
	
	int tx = get_local_id (0) ; // 2D local ID x
	int ty = get_local_id (1) ; // 2D local ID y

	// first and last sub-matrix of A for this block
	int aBegin = wA * BLOCK_SIZE * by ;
	int aEnd = aBegin + wA - 1 ;
	int aStep = BLOCK_SIZE ;

	// first and last sub-matrix of B for this block
	int bBegin = BLOCK_SIZE * bx ;
	int bStep = BLOCK_SIZE * wB;
	
	float Csub = 0.0 ;
	// Iterate over all sub-matrices of A and B
	
	for (int a = aBegin , b = bBegin ; a <= aEnd ; a+=aStep, b+=bStep) {
	
		// Static work-group local allocations
		__local float As[BLOCK_SIZE][BLOCK_SIZE] ;
		__local float Bs[BLOCK_SIZE][BLOCK_SIZE] ;
		

		// Each thread loads one element of the block
		// from global memory
		As[ty][tx] = A[ a + wA * ty + tx] ;
		Bs[ty][tx] = B[ b + wB * ty + tx] ;

		// Barrier to synchronize all threads
		barrier (CLK_LOCAL_MEM_FENCE) ;
		// Now the local sub-matricies As and Bs are valid

		// Multiply the two sub-matrices. Each thread
		// computes one element of the block sub-ma t r i x .
		for ( int k = 0 ; k < BLOCK_SIZE ; ++k )
			Csub += As[ty][k] * Bs[k][tx] ;

		// Barrier to synchronize all  threads before moving
		barrier (CLK_LOCAL_MEM_FENCE) ;

	}

	int c = wB * BLOCK_SIZE * by + BLOCK_SIZE * bx ;
	C[ c + wB * ty + tx ] = Csub ; // write to global memory
}


